import asyncio
import json
import logging
from datetime import datetime, timezone
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp
import cbor2

import numpy as np
import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.event import InternalWorkerMessage, message_tag_hook
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.skipif("config.getoption('rust')", reason="rust does not support dumping")
@pytest.mark.asyncio
async def test_dump(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_orca: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_small: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    tmp_path: Any,
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_worker(WorkerName("w2"))
    await create_worker(WorkerName("w3"))

    p_eiger = tmp_path / "eiger_dump.cbors"
    print(p_eiger, type(p_eiger))

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
                dump_path=p_eiger,
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("orca")],
                upstream_url=Url("tcp://localhost:9998"),
                ingester_url=Url("tcp://localhost:10011"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("alba")],
                upstream_url=Url("tcp://localhost:9997"),
                ingester_url=Url("tcp://localhost:10012"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("slow")],
                upstream_url=Url("tcp://localhost:9996"),
                ingester_url=Url("tcp://localhost:10013"),
            ),
        )
    )

    await wait_for_controller(
        streams={
            StreamName("eiger"),
            StreamName("orca"),
            StreamName("alba"),
            StreamName("slow"),
        }
    )
    async with aiohttp.ClientSession() as session:
        p_prefix = tmp_path / "dump_"
        logging.info("prefix is %s encoded: %s", p_prefix, str(p_prefix).encode("utf8"))

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/dump_prefix",
            data=str(p_prefix).encode("utf8"),
        )
        assert resp.status == 200

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
                "orca": [
                    [
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json")
                    ]
                    for i in range(1, ntrig)
                ],
                "alba": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json"),
                    ]
                    for i in range(1, ntrig)
                ],
                "slow": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json"),
                    ]
                    if i % 4 == 0
                    else None
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))
    asyncio.create_task(stream_orca(context, 9998, ntrig - 1))
    asyncio.create_task(stream_small(context, 9997, ntrig - 1))
    asyncio.create_task(stream_small(context, 9996, ntrig // 4))

    content = await wait_for_finish()

    assert content == {
        "last_assigned": ntrig + 1,
        "completed_events": ntrig + 1,
        "total_events": ntrig + 1,
        "finished": True,
    }

    # read dump
    with open(p_eiger, "rb") as f:
        evs = []
        while True:
            try:
                dat = cbor2.load(f, tag_hook=message_tag_hook)
                if isinstance(dat, InternalWorkerMessage):
                    evs.append(dat.event_number)
            except EOFError:
                break

    assert evs == list(range(0, ntrig + 1))

    # read prefix dump
    with open(f"{p_prefix}orca-ingester-{uuid}.cbors", "rb") as f:
        evs = []
        while True:
            try:
                before = datetime.now(timezone.utc)
                dat = cbor2.load(f, tag_hook=message_tag_hook)
                if isinstance(dat, InternalWorkerMessage):
                    evs.append(dat.event_number)
                    assert (
                        dat.created_at < before
                    ), "timestamp was not created when loading the data from file but read"
            except EOFError:
                break

    assert evs == list(range(0, ntrig + 1))

    context.destroy()


@pytest.mark.skipif("config.getoption('rust')", reason="rust does not support dumping")
@pytest.mark.asyncio
async def test_dump_xrd(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_small_xrd: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    tmp_path: Any,
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("xrd")],
                upstream_url=Url("tcp://localhost:9999"),
                ingester_url=Url("tcp://localhost:10010"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("xrd")})
    async with aiohttp.ClientSession() as session:
        p_prefix = tmp_path / "dump_"
        logging.info("prefix is %s encoded: %s", p_prefix, str(p_prefix).encode("utf8"))

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/dump_prefix",
            data=str(p_prefix).encode("utf8"),
        )
        assert resp.status == 200

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "xrd": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_small_xrd(context, 9999, ntrig - 1))

    content = await wait_for_finish()

    assert content == {
        "last_assigned": ntrig + 1,
        "completed_events": ntrig + 1,
        "total_events": ntrig + 1,
        "finished": True,
    }
    context.destroy()


@pytest.mark.skipif("config.getoption('rust')", reason="rust does not support dumping")
@pytest.mark.asyncio
async def test_dump_map_and_parameters(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    tmp_path: Any,
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("xrd")],
                upstream_url=Url("tcp://localhost:9999"),
                ingester_url=Url("tcp://localhost:10010"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("xrd")})
    async with aiohttp.ClientSession() as session:
        p_prefix = tmp_path / "dump_"
        logging.info("prefix is %s encoded: %s", p_prefix, str(p_prefix).encode("utf8"))

        pars = {
            "dump_prefix": str(p_prefix).encode("utf8"),
            "test_string": "spam".encode("utf8"),
            "test_int": "42".encode("utf8"),
        }
        for name, data in pars.items():
            resp = await session.post(
                f"http://localhost:5000/api/v1/parameter/{name}",
                data=data,
            )
            assert resp.status == 200

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "xrd": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

    context = zmq.asyncio.Context()

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/mapping")
        mapping = await st.json()
        logging.debug("mapping %s", mapping)

        with open(f"{p_prefix}mapping-{uuid}.json", "rb") as f:
            # logging.info(f"Content of mapping {f.read()}")
            dumped_mapping = json.load(f)
            logging.debug("dumped_mapping %s", dumped_mapping)
        assert mapping == dumped_mapping

    with open(f"{p_prefix}parameters-{uuid}.json", "rb") as f:
        dumped_pars = json.load(f)
        logging.debug("dumped_pars %s", dumped_pars)
        for p in dumped_pars:
            assert p["data"].encode("utf8") == pars[p["name"]]

    context.destroy()


@pytest.mark.skipif("config.getoption('rust')", reason="rust does not support dumping")
@pytest.mark.asyncio
async def test_dump_bin_parameters(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    tmp_path: Any,
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("xrd")],
                upstream_url=Url("tcp://localhost:9999"),
                ingester_url=Url("tcp://localhost:10010"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("xrd")})
    async with aiohttp.ClientSession() as session:
        p_prefix = tmp_path / "dump_"
        logging.info("prefix is %s encoded: %s", p_prefix, str(p_prefix).encode("utf8"))

        pars = {
            "dump_prefix": str(p_prefix).encode("utf8"),
            "test_null": np.zeros((2, 2)).tobytes(),
        }
        for name, data in pars.items():
            resp = await session.post(
                f"http://localhost:5000/api/v1/parameter/{name}",
                data=data,
            )
            assert resp.status == 200

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "xrd": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

    context = zmq.asyncio.Context()

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/mapping")
        mapping = await st.json()
        logging.debug("mapping %s", mapping)

        with open(f"{p_prefix}mapping-{uuid}.json", "rb") as f:
            # logging.info(f"Content of mapping {f.read()}")
            dumped_mapping = json.load(f)
            logging.debug("dumped_mapping %s", dumped_mapping)
        assert mapping == dumped_mapping

    with open(f"{p_prefix}parameters-{uuid}.cbor", "rb") as f:
        pars_dict = {}
        dumped_pars = cbor2.load(f)
        logging.debug("dumped_pars %s", dumped_pars)
        for p in dumped_pars:
            pars_dict[p["name"]] = p["data"]
        assert pars_dict["dump_prefix"].encode("utf8") == pars["dump_prefix"]
        assert pars_dict["test_null"] == pars["test_null"]

    context.destroy()



from datetime import datetime, timezone
from typing import Any, Optional

import zmq
from cbor2 import CBORTag, CBOREncoder, CBORDecoder

from dranspose.protocol import EventNumber, StreamName, WorkerName, HashDigest
from pydantic import BaseModel, ConfigDict, computed_field, Field


class StreamData(BaseModel):
    """
    Data container for a single stream with all zmq frames belonging to it

    Attributes:
         typ: arbitrary typ set by the ingester to common parsing
         frames: all frames received for this event for the stream
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    typ: str
    frames: list[zmq.Frame] | list[bytes]

    @computed_field  # type: ignore[prop-decorator]
    @property
    def length(self) -> int:
        """
        Calculates the length of frames.

        Returns:
             length of frames
        """
        return len(self.frames)

    def get_bytes(self) -> "StreamData":
        """
        Copies the data from the zmq buffer

        Returns:
             An object with a list of bytes.
        """
        return StreamData(
            typ=self.typ,
            frames=[
                frame.bytes if isinstance(frame, zmq.Frame) else frame
                for frame in self.frames
            ],
        )


class InternalWorkerMessage(BaseModel):
    """
    A container for partial events which carries one or more streams. This is the message between ingesters and workers.

    Attributes:
        event_number: event number
        streams: one or more streams from an ingester
    """

    event_number: EventNumber
    streams: dict[StreamName, StreamData] = {}
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))

    def get_all_frames(self) -> list[zmq.Frame | bytes]:
        return [frame for stream in self.streams.values() for frame in stream.frames]


def message_encoder(encoder: CBOREncoder, value: Any) -> None:
    # Tag number 4000 was chosen arbitrarily
    if isinstance(value, InternalWorkerMessage):
        encoder.encode(
            CBORTag(42877, (value.event_number, value.streams, value.created_at))
        )
    elif isinstance(value, StreamData):
        encoder.encode(CBORTag(42878, (value.typ, value.frames)))
    else:
        raise TypeError(type(value).__name__)


def message_tag_hook(
    decoder: CBORDecoder, tag: CBORTag, shareable_index: Any = None
) -> Any:
    if tag.tag == 42877:
        if len(tag.value) == 3:
            return InternalWorkerMessage(
                event_number=tag.value[0], streams=tag.value[1], created_at=tag.value[2]
            )
        return InternalWorkerMessage(event_number=tag.value[0], streams=tag.value[1])
    elif tag.tag == 42878:
        return StreamData(typ=tag.value[0], frames=tag.value[1])
    else:
        return tag


class ResultData(BaseModel):
    """
    Container for transferring results from the worker to the reducer. In enhances the pure payload with useful meta data

    Attributes:
        event_number: which event the result belongs to. NB: results may arrive out of order
        worker: which worker processed it.
        parameters_hash: which version of parameters was used to process the event
        payload: the data return from the custom worker function: process_event
    """

    event_number: EventNumber
    worker: WorkerName
    parameters_hash: Optional[HashDigest]
    payload: Any


class EventData(BaseModel):
    """
    Main container for an event provided to the worker function

    Attributes:
        event_number:    Current event number relative to the trigger map provided
        streams:         Data for each stream present in the event
    """

    event_number: EventNumber
    streams: dict[StreamName, StreamData]

    @classmethod
    def from_internals(cls, msgs: list[InternalWorkerMessage]) -> "EventData":
        """
        Helper function to assemble an event from the internal messages received from the ingesters.
        This is a factory method

        Args:
             msgs: Internal messages each containing a subset of the streams for the event

        Returns:
            A new object with all data combined.
        """
        assert len(msgs) > 0, "merge at least one message"
        assert (
            len(set([m.event_number for m in msgs])) == 1
        ), f"Cannot merge data from events {[m.event_number for m in msgs]}"
        all_stream_names = [stream for m in msgs for stream in m.streams.keys()]
        assert len(all_stream_names) == len(
            set(all_stream_names)
        ), "Cannot merge data with duplicate streams"

        ret = EventData(event_number=msgs[0].event_number, streams={})
        for msg in msgs:
            ret.streams.update(msg.streams)
        return ret



import logging
import os
import pathlib
import pickle
from pathlib import PosixPath
import h5pyd

import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional
import zmq.asyncio

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqsub_contrast import (
    ZmqSubContrastIngester,
    ZmqSubContrastSettings,
)
from dranspose.ingesters.zmqsub_xspress3 import (
    ZmqSubXspressIngester,
    ZmqSubXspressSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_finish, wait_for_controller


@pytest.mark.asyncio
async def test_reduction(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
    tmp_path: pathlib.PurePath,
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.dummy.worker:FluorescenceWorker",
            ),
        )
    )
    p_contrast = tmp_path / "contrast_ingest.pkls"

    await create_ingester(
        ZmqSubContrastIngester(
            settings=ZmqSubContrastSettings(
                ingester_streams=[StreamName("contrast")],
                upstream_url=Url("tcp://localhost:5556"),
                ingester_url=Url("tcp://localhost:10000"),
                dump_path=p_contrast,
            ),
        )
    )
    p_xspress = tmp_path / "xspress_ingest.pkls"
    await create_ingester(
        ZmqSubXspressIngester(
            settings=ZmqSubXspressSettings(
                ingester_streams=[StreamName("xspress3")],
                upstream_url=Url("tcp://localhost:9999"),
                ingester_url=Url("tcp://localhost:10001"),
                dump_path=p_xspress,
            ),
        )
    )

    await wait_for_controller(streams={StreamName("contrast"), StreamName("xspress3")})
    async with aiohttp.ClientSession() as session:
        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/roi1",
            json=[0, 10],
        )
        assert resp.status == 200
        await resp.json()

        ntrig = 20
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "contrast": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig)
                ],
                "xspress3": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(
        stream_pkls(
            context,
            9999,
            PosixPath("tests/data/xspress3mini-dump20.pkls"),
            0.001,
            zmq.PUB,
        )
    )
    asyncio.create_task(
        stream_pkls(
            context, 5556, PosixPath("tests/data/contrast-dump.pkls"), 0.001, zmq.PUB
        )
    )

    content = await wait_for_finish()
    assert content == {
        "last_assigned": ntrig + 2,
        "completed_events": ntrig + 2,
        "total_events": ntrig + 2,
        "finished": True,
    }

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5001/api/v1/result/pickle")
        content_bytes = await st.content.read()
        result = pickle.loads(content_bytes)
        assert result == {
            "map": {
                (-2.004051863, -2.002903037): {"roi1": 81},
                (-2.001404022, -0.9992748245): {"roi1": 76},
                (-2.000941193, 0.001238407776): {"roi1": 67},
                (-2.000656853, 1.001692484): {"roi1": 49},
                (-2.000562814, 2.0019763): {"roi1": 69},
                (-0.6658021229, -2.002587864): {"roi1": 97},
                (-0.6664982812, -0.9992643584): {"roi1": 75},
                (-0.6666898134, 0.001223421204): {"roi1": 58},
                (-0.6666922578, 1.001631271): {"roi1": 81},
                (-0.6666616732, 2.001930725): {"roi1": 77},
                (0.6676852429, -2.002412156): {"roi1": 91},
                (0.6666162969, -0.9992055423): {"roi1": 81},
                (0.6664269385, 0.001214749016): {"roi1": 76},
                (0.6663550388, 1.00170782): {"roi1": 67},
                (0.6663102034, 2.002018839): {"roi1": 49},
                (2.001407847, -2.002363463): {"roi1": 69},
                (2.000183463, -0.9992309104): {"roi1": 97},
                (1.999963432, 0.001258902382): {"roi1": 75},
                (1.999884878, 1.001690624): {"roi1": 58},
                (1.999828501, 2.001959397): {"roi1": 81},
            },
            "version": {
                "commit_hash": "1a474d1455f0d6c6113a22179b8fff1e98325916",
                "branch_name": "main",
                "timestamp": "2024-04-25T08:07:59+02:00",
                "repository_url": "https://gitlab.maxiv.lu.se/DanMAX/pipeline/drp-tomo-buffer",
            },
        }
        assert len(result["map"].keys()) == ntrig
        assert (
            result["version"]["commit_hash"]
            == "1a474d1455f0d6c6113a22179b8fff1e98325916"
        )

    def work() -> None:
        f = h5pyd.File("http://localhost:5001/", "r")
        logging.info("file %s", list(f.keys()))
        logging.info("map %s", list(f["map"].keys()))
        logging.info("version %s", f["version/commit_hash"][()])
        assert (
            f["version"]["commit_hash"][()]
            == b"1a474d1455f0d6c6113a22179b8fff1e98325916"
        )

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work)

    context.destroy()



import numpy as np

from dranspose.helpers.jsonpath_slice_ext import NumpyExtentedJsonPathParser


def test_jsonpath() -> None:
    res = NumpyExtentedJsonPathParser(debug=False).parse("map, meta")
    print(res.__repr__())
    dat = {"map": 1, "meta": [1, 2, 3]}
    result = res.find(dat)
    print("or result", [r.value for r in result])
    res = NumpyExtentedJsonPathParser(debug=False).parse("map.`numpy(:,6:8,5)`")
    print(res.__repr__())

    data = {"map": np.linspace(0, 999, 1000).reshape((10, 10, 10)), "meta": [1, 2, 3]}
    # print(data)

    result = [r.value for r in res.find(data)]
    print(result[0])
    arr = np.array(
        [
            [65.0, 75.0],
            [165.0, 175.0],
            [265.0, 275.0],
            [365.0, 375.0],
            [465.0, 475.0],
            [565.0, 575.0],
            [665.0, 675.0],
            [765.0, 775.0],
            [865.0, 875.0],
            [965.0, 975.0],
        ]
    )
    print(arr)
    assert arr.tolist() == result[0].tolist()



import json

import zmq

from dranspose.event import InternalWorkerMessage, StreamData
from dranspose.protocol import EventNumber, StreamName


def test_header_serialisation() -> None:
    orca = StreamData(typ="STINS", frames=[zmq.Frame(b"asd")])
    lam = StreamData(
        typ="STINS", frames=[zmq.Frame(b"lam1"), zmq.Frame(b"lam2"), zmq.Frame(b"lam3")]
    )
    alba = StreamData(typ="STINS", frames=[zmq.Frame(b"alba")])
    message = InternalWorkerMessage(event_number=EventNumber(42))
    message.streams[StreamName("orca")] = orca
    message.streams[StreamName("lambda")] = lam
    message.streams[StreamName("alba")] = alba
    dump = message.model_dump_json(
        exclude={"streams": {"__all__": "frames"}}
    )  #: {"frames"}})

    buffers = message.get_all_frames()
    assert [b"asd", b"lam1", b"lam2", b"lam3", b"alba"] == [
        b.bytes if isinstance(b, zmq.Frame) else b for b in buffers
    ]
    assert (
        dump[:148]
        == '{"event_number":42,"streams":{"orca":{"typ":"STINS","length":1},"lambda":{"typ":"STINS","length":3},"alba":{"typ":"STINS","length":1}},"created_at":'
    )
    prelim = json.loads(dump)
    assert prelim == {
        "event_number": 42,
        "streams": {
            "orca": {"typ": "STINS", "length": 1},
            "lambda": {"typ": "STINS", "length": 3},
            "alba": {"typ": "STINS", "length": 1},
        },
        "created_at": prelim["created_at"],
    }
    pos = 0
    for stream, data in prelim["streams"].items():
        print(stream, data)
        data["frames"] = buffers[pos : pos + data["length"]]
        assert data["typ"] == "STINS"
        pos += data["length"]
    assert pos == len(buffers)
    msg = InternalWorkerMessage.model_validate(prelim)
    assert msg.event_number == 42

    albaframe = msg.streams[StreamName("alba")].frames[0]
    if isinstance(albaframe, zmq.Frame):
        albaframe = albaframe.bytes
    lamframe = msg.streams[StreamName("lambda")].frames[2]
    if isinstance(lamframe, zmq.Frame):
        lamframe = lamframe.bytes
    orcaframe = msg.streams[StreamName("orca")].frames[0]
    if isinstance(orcaframe, zmq.Frame):
        orcaframe = orcaframe.bytes
    assert albaframe == b"alba"
    assert lamframe == b"lam3"
    assert orcaframe == b"asd"


def test_out_of_band_frames() -> None:
    buffers = [zmq.Frame(b"asd")]
    orca = StreamData(typ="STINS", frames=buffers)

    dump = orca.model_dump_json(exclude={"frames"})

    assert dump == '{"typ":"STINS","length":1}'

    prelim = json.loads(dump)
    assert prelim == {"typ": "STINS", "length": 1}
    prelim["frames"] = buffers
    sd = StreamData.model_validate(prelim)
    assert sd.frames == buffers



import asyncio
import logging
import multiprocessing
from typing import Any

import aiohttp
import pytest
import redis.asyncio as redis
import uvicorn
from dranspose.protocol import (
    RedisKeys,
    EnsembleState,
)


class UvicornServer(multiprocessing.Process):
    def __init__(self, config: uvicorn.Config):
        super().__init__()
        self.server = uvicorn.Server(config=config)
        self.config = config

    def stop(self) -> None:
        self.terminate()

    def run(self, *args: Any, **kwargs: Any) -> None:
        self.server.run()


# A lot of distributed state assumes that there is only ever one controller with the same prefix running per redis instance
# This does not hold when redeploying on k8s as it starts the new container before the old is terminated.
# The controller now implements a lock which delays the start of a controller if another is still holding it and wait to properly finish.
# An artifact of interleaved stopping and starting was, that the parameters created by the new controller were cleaned up by the closing of the old one and then led to inconsistent parameters.
# In addition to the locking of the controller, the consistent_parameters routine assures that all parameters with values are present in redis.


@pytest.mark.asyncio
async def test_restart() -> None:
    config = uvicorn.Config(
        "dranspose.controller:app", host="127.0.0.1", port=5000, log_level="debug"
    )
    instance = UvicornServer(config=config)
    instance.start()

    r = redis.Redis(host="localhost", port=6379, decode_responses=True, protocol=3)

    async with aiohttp.ClientSession() as session:
        running = False
        while not running:
            await asyncio.sleep(0.3)
            try:
                st = await session.get("http://localhost:5000/api/v1/config")
                running = True
            except aiohttp.client_exceptions.ClientConnectorError:
                pass

        state = EnsembleState.model_validate(await st.json())
        logging.info("ensemble state uuids are %s", state)

        resp = await session.post(
            "http://localhost:5000/api/v1/parameter/testname",
            data=b"testvalue",
        )
        assert resp.status == 200

        lock = await r.keys(RedisKeys.lock())
        logging.info("lock is %s", lock)

        params = await r.keys(RedisKeys.parameters("*", "*"))
        logging.info("redis keys are %s", params)

        assert len(params) == 1
        assert params[0].startswith("dranspose:parameters:testname:")

        config2 = uvicorn.Config(
            "dranspose.controller:app", host="127.0.0.1", port=5001, log_level="debug"
        )
        instance2 = UvicornServer(config=config2)
        instance2.start()

        instance.stop()

        running = False
        while not running:
            await asyncio.sleep(0.3)
            try:
                st = await session.get("http://localhost:5001/api/v1/config")
                running = True
            except aiohttp.client_exceptions.ClientConnectorError:
                pass

        state = EnsembleState.model_validate(await st.json())
        logging.info("new controller state %s", state)

        resp = await session.post(
            "http://localhost:5001/api/v1/parameter/testnew",
            data=b"newvalue",
        )
        assert resp.status == 200

        params = await r.keys(RedisKeys.parameters("*", "*"))
        logging.info("redis keys are %s", params)

        assert any([p.startswith("dranspose:parameters:testnew") for p in params])

        running = True
        while running:
            await asyncio.sleep(0.3)
            try:
                await session.get("http://localhost:5000/api/v1/config")
            except aiohttp.client_exceptions.ClientConnectorError:
                running = False

    params = await r.keys(RedisKeys.parameters("*", "*"))
    logging.info("redis keys after old stop %s", params)

    assert any([p.startswith("dranspose:parameters:testnew") for p in params])

    instance2.stop()

    await r.aclose()



import asyncio
import pickle
from datetime import datetime, timezone, timedelta
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp
import numpy as np

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.data.stream1 import Stream1Data
from dranspose.event import EventData
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.middlewares.stream1 import parse
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    WorkerTag,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_simple(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    debug_worker: Callable[[Optional[str], Optional[list[str]]], Awaitable[None]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await debug_worker("debugworker", ["debug"])
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    []
                    if 3 < i < 8
                    else [
                        VirtualWorker(tags={WorkerTag("debug")}).model_dump(mode="json")
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    await wait_for_finish()

    context.destroy()

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5002/api/v1/last_events?number=20")
        content = await st.content.read()
        res: list[EventData] = pickle.loads(content)
        got_frames: list[int] = []
        for ev in res:
            assert set(ev.streams.keys()) == {"eiger"}
            pkt = parse(ev.streams[StreamName("eiger")])
            if isinstance(pkt, Stream1Data):
                assert pkt.frame == ev.event_number - 1
                assert pkt.shape == [1475, 831]
                assert pkt.type == "uint16"
                assert pkt.compression == "none"
                assert "dummy" in pkt.timestamps
                assert (
                    datetime.now(timezone.utc) - timedelta(seconds=60)
                    < datetime.fromisoformat(pkt.timestamps["dummy"])
                    < datetime.now(timezone.utc)
                )
                assert pkt.data.dtype == np.uint16
                got_frames.append(pkt.frame)
    assert got_frames == list(range(0, 3)) + list(range(7, 9))



from typing import Callable, Awaitable

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    VirtualWorker,
    VirtualConstraint,
    StreamName,
)
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_status(controller: None) -> None:
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/status")
        await st.json()
        assert st.status == 200


@pytest.mark.asyncio
async def test_stream_not_available(controller: None) -> None:
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/status")
        assert st.status == 200
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 400
        response = await resp.json()
        assert {"detail": "streams {'eiger'} not available"} == response


@pytest.mark.asyncio
async def test_not_enough_workers(
    controller: None, create_ingester: Callable[[Ingester], Awaitable[Ingester]]
) -> None:
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 400
        response = await resp.json()
        assert {"detail": "only 0 workers available, but 1 required"} == response
